//
// Copyright (C) 2004, 2005 Pingtel Corp.
// 
//
// $$
////////////////////////////////////////////////////////////////////////

#ifndef _UtlHistogram_h_
#define _UtlHistogram_h_

// SYSTEM INCLUDES

// APPLICATION INCLUDES
#include "utl/UtlString.h"

// DEFINES
// MACROS
// EXTERNAL FUNCTIONS
// EXTERNAL VARIABLES
// CONSTANTS
// STRUCTS
// TYPEDEFS
// FORWARD DECLARATIONS

/**
 * Record and dump counts in a series of bins.
 *
 * Recorded values are integers.  The histogram has a specified number
 * of bins, each counting the number of values over a range, each of
 * which has the same specified width.
 *
 * The number of bins implemented is 2 more than the specified number,
 * to allow one bin for values lower than the range and one for values
 * higher than the range.
 *
 * The minimum of the range of bin 0 is Base, and each normal bin has
 * a range of Size values.
 *
 * The first normal bin is numbered 0, so the bin for too-low values
 * is numbered -1.  Thus, the bins record:
 *
 * Bin -1 counts values less than "Base".
 * Bin 0 counts values from "Base" to "Base + Size - 1".
 * Bin 1 counts values from "Base + Size" to "Base + 2*Size - 1".
 * Bin n counts values from "Base + n*Size" to "Base + (n+1)*Size - 1".
 * Bin NoOfBins-1 counts values from "Base + (NoOfBins-1)*Size" to
 *         "Base + NoOfBins*Size -1".
 * Bin NoOfBins counts values greater than or equal to "Base + NoOfBins*Size".
 *
 * The values of all the bins can be extracted in a string by setting
 * outputFormat and outputWidth when the histogram is created.
 * Calling show() formats each bin's value via
 * sprintf(buffer, outputFomat, bin-value), concatenates them
 * together, and returns the result as a UtlString.
 * outputFormat must be a format string for formatting a single sipx_int_t,
 * and it must always generate at most outputWidth characters.
 *
 * The outputFormat string must remain valid as sipx_long_t as the histogram
 * object exists; the returned UtlString must be freed by the caller.
 */
class UtlHistogram
{
/* //////////////////////////// PUBLIC //////////////////////////////////// */
  public:

/* ============================ CREATORS ================================== */

   /**
    * Constructor.
    */
   UtlHistogram(sipx_uint_t bins, sipx_int_t base, sipx_uint_t size,
                const char* outputFormat = "", sipx_uint_t outputWidth = 0);
     
   /**
    * Destructor
    */
   ~UtlHistogram();

/* ============================ MANIPULATORS ============================== */

   /**
    * Record a value.
    * Returns the number of counts in the histogram.
    */
   sipx_uint_t tally(sipx_int_t);

   /**
    * Clear the histogram.
    */
   void clear();

/* ============================ ACCESSORS ================================= */

   /**
    * Get the specified number of bins (which is 2 less than the total number
    * of bins).
    */
   sipx_uint_t getNoOfBins();

   /**
    * Get the lowest value for bin 0.
    */
   sipx_uint_t getBase();

   /**
    * Get the size of each bin.
    */
   sipx_uint_t getBinSize();

   /**
    * Get the total count.
    */
   sipx_uint_t getCount();

   /**
    * Get the count in bin i.
    * i ranges from -1 to NoOfBins+1.
    */
   sipx_uint_t operator[](sipx_uint_t i);

/* ============================ INQUIRY =================================== */

   /**
    * Get a string containing the formatted values from the bins.
    * The caller must free the returned UtlString.
    */
   UtlString* show();

/* //////////////////////////// PROTECTED ///////////////////////////////// */
  protected:

/* //////////////////////////// PRIVATE /////////////////////////////////// */
  private:

   // Number of bins.
   sipx_uint_t mNoOfBins;
   // Lowest value in bin 0.
   sipx_int_t mBase;
   // Size of each bin:
   sipx_uint_t mBinSize;
   // Posipx_int_ter to an array of mNoOfBins sipx_int_t's.
   sipx_uint_t* mpBins;
   // Total number of counts.
   sipx_uint_t mCount;

   // Format used to output values.
   const char* mOutputFormat;
   // Width of output that format will generate.
   sipx_uint_t mOutputWidth;
};

/* ============================ INLINE METHODS ============================ */

#endif  // _UtlHistogram_h_
