//
// Copyright (C) 2004, 2005 Pingtel Corp.
// 
//
// $$
////////////////////////////////////////////////////////////////////////

// SYSTEM INCLUDES
#include <assert.h>
#include <stdio.h>

#if defined(_WIN32)
#   include <winsock.h>
#elif defined(_VXWORKS)
#   include <inetLib.h>
#   include <netdb.h>
#   include <resolvLib.h>
#   include <sockLib.h>
#elif defined(__pingtel_on_posix__)
#   include <sys/types.h>
#   include <sys/socket.h>
#   include <netdb.h>
#   include <netinet/in.h>
#   include <arpa/inet.h>
#else
#error Unsupported target platform.
#endif

// APPLICATION INCLUDES

//SSL related includes
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#include "os/OsSSL.h"
#include "os/OsSSLConnectionSocket.h"
#include "os/OsUtil.h"
#include "os/OsSysLog.h"
#include "os/OsTask.h"

// EXTERNAL FUNCTIONS
// EXTERNAL VARIABLES

// CONSTANTS

// STATIC VARIABLE INITIALIZATIONS

/* //////////////////////////// PUBLIC //////////////////////////////////// */

/* ============================ CREATORS ================================== */

// Constructor
OsSSLConnectionSocket::OsSSLConnectionSocket(sipx_int_t connectedSocketDescriptor, sipx_long_t timeoutInSecs) :
    OsConnectionSocket(NULL, connectedSocketDescriptor),
    mSSL(NULL)
{
   if (mIsConnected)
   {
      mbExternalSSLSocket = FALSE;
      SSLInitSocket(connectedSocketDescriptor, timeoutInSecs);
      OsSysLog::add(FAC_KERNEL, PRI_DEBUG, 
                    "OsSSLConnectionSocket::_(socket %d , timeout %ld)",
                    connectedSocketDescriptor, timeoutInSecs
                    );
   }
   else
   {
      OsSysLog::add(FAC_KERNEL, PRI_ERR, 
                    "OsSSLConnectionSocket::_ underlying OsConnectionSocket not connected!!");
   }
}

OsSSLConnectionSocket::OsSSLConnectionSocket(SSL *s, sipx_int_t connectedSocketDescriptor) :
    OsConnectionSocket(NULL, connectedSocketDescriptor)
{
    mbExternalSSLSocket = TRUE;
    mSSL = s;
    socketDescriptor = connectedSocketDescriptor;
    OsSysLog::add(FAC_KERNEL, PRI_DEBUG, 
                  "OsSSLConnectionSocket::_(SSL %p, socket %d)", s, connectedSocketDescriptor);
}

// Constructor
OsSSLConnectionSocket::OsSSLConnectionSocket(sipx_int_t serverPort, const char* serverName,
                                             sipx_long_t timeoutInSecs) :
    OsConnectionSocket(serverPort,serverName),
    mSSL(NULL)
{
    mbExternalSSLSocket = FALSE;
    if (mIsConnected)
    {
       SSLInitSocket(socketDescriptor, timeoutInSecs);
       OsSysLog::add(FAC_KERNEL, PRI_DEBUG, 
                     "OsSSLConnectionSocket::_(port %d, name '%s', timeout %ld)",
                     serverPort, serverName, timeoutInSecs
                     );
    }
}


// Destructor
OsSSLConnectionSocket::~OsSSLConnectionSocket()
{
   OsSysLog::add(FAC_KERNEL, PRI_DEBUG, 
                 "OsSSLConnectionSocket::~"
                 );
   remoteHostName = OsUtil::NULL_OS_STRING;
   close();
}

/* ============================ MANIPULATORS ============================== */

// Assignment operator
OsSSLConnectionSocket&
OsSSLConnectionSocket::operator=(const OsSSLConnectionSocket& rhs)
{
   if (this == &rhs)            // handle the assignment to self case
      return *this;

   return *this;
}

UtlBoolean OsSSLConnectionSocket::reconnect()
{
    return FALSE; //TODO
}

void OsSSLConnectionSocket::close()
{
    if (mSSL)
    {
       if (mIsConnected)
          SSL_shutdown(mSSL);
       else
          SSL_clear(mSSL);
    }

   /* Clean up. */
    OsConnectionSocket::close();

   if (mSSL)
   {
      SSL_free(mSSL);
      ERR_remove_state(0);
      mSSL = NULL;
   }
}

sipx_int_t OsSSLConnectionSocket::write(const char* buffer, sipx_int_t bufferLength)
{
    sipx_int_t bytesWritten = -1;

    bytesWritten = SSL_write (mSSL, buffer, bufferLength);

    return(bytesWritten);
}

sipx_int_t OsSSLConnectionSocket::write(const char* buffer, sipx_int_t bufferLength, sipx_long_t waitMilliseconds)
{
    sipx_int_t bytesWritten = -1;
    if(isReadyToWrite(waitMilliseconds))
    {
        bytesWritten = SSL_write (mSSL, buffer, bufferLength);
    }

    return(bytesWritten);
}

sipx_int_t OsSSLConnectionSocket::read(char* buffer, sipx_int_t bufferLength)
{
    // Use base class implementation
    sipx_int_t bytesRead = -1;

    bytesRead = SSL_read (mSSL, buffer, bufferLength);

#ifdef VALGRIND_MAKE_READABLE
    // If we are using Valgrind, we have to compensate for the fact that
    // Valgrind thinks all the output of SSL_read is undefined.
    // (This probably comes from SSL's use of uninitialized memory as part
    // of its key-generation algorithm.)
    VALGRIND_DISCARD(VALGRIND_MAKE_READABLE(&bytesRead, sizeof (bytesRead)));
    if (bytesRead > 0)
    {
       VALGRIND_DISCARD(VALGRIND_MAKE_READABLE(buffer, bytesRead));
    }
#endif /* VALGRIND_MAKE_READABLE */

    return(bytesRead);
}

sipx_int_t OsSSLConnectionSocket::read(char* buffer,
                             sipx_int_t bufferLength,
                             UtlString* ipAddress,
                             sipx_int_t* port)
{
    // Overide base class version as recvfrom does not
    // seem to return host info correctly for TCP
    // Use base class version without the remote host info
    sipx_int_t bytesRead = -1;

    bytesRead = SSL_read (mSSL, buffer, bufferLength);

#ifdef VALGRIND_MAKE_READABLE
    // If we are using Valgrind, we have to compensate for the fact that
    // Valgrind thinks all the output of SSL_read is undefined.
    // (This probably comes from SSL's use of uninitialized memory as part
    // of its key-generation algorithm.)
    VALGRIND_DISCARD(VALGRIND_MAKE_READABLE(&bytesRead, sizeof (bytesRead)));
    if (bytesRead > 0)
    {
       VALGRIND_DISCARD(VALGRIND_MAKE_READABLE(buffer, bytesRead));
    }
#endif /* VALGRIND_MAKE_READABLE */
    // Explicitly get the remote host info.
    getRemoteHostIp(ipAddress);
    *port = getRemoteHostPort();
    return(bytesRead);
}

// Because we have overided one read method, we
// must implement them all in OsSSLConnectionSocket or
// we end up hiding some of the methods.
sipx_int_t OsSSLConnectionSocket::read(char* buffer,
                            sipx_int_t bufferLength,
                            sipx_long_t waitMilliseconds)
{

    sipx_int_t bytesRead = -1;
    if(isReadyToRead(waitMilliseconds))
    {
        bytesRead = SSL_read (mSSL, buffer, bufferLength);
    }
#ifdef VALGRIND_MAKE_READABLE
    // If we are using Valgrind, we have to compensate for the fact that
    // Valgrind thinks all the output of SSL_read is undefined.
    // (This probably comes from SSL's use of uninitialized memory as part
    // of its key-generation algorithm.)
    VALGRIND_DISCARD(VALGRIND_MAKE_READABLE(&bytesRead, sizeof (bytesRead)));
    if (bytesRead > 0)
    {
       VALGRIND_DISCARD(VALGRIND_MAKE_READABLE(buffer, bytesRead));
    }
#endif /* VALGRIND_MAKE_READABLE */
    return(bytesRead);
}

/* ============================ ACCESSORS ================================= */
/* ============================ INQUIRY =================================== */
OsSocket::IpProtocolSocketType OsSSLConnectionSocket::getIpProtocol() const
{
    return(OsSocket::SSL_SOCKET);
}

/// Is this connection encrypted using TLS/SSL?
bool OsSSLConnectionSocket::isEncrypted() const
{
   return false;
}

   
/// Get any authenticated peer host names.
bool OsSSLConnectionSocket::peerIdentity( UtlSList* altNames
                                         ,UtlString* commonName
                                         ) const
{
   /*
    * - true if the connection is TLS/SSL and the peer has presented
    *        a certificate signed by a trusted certificate authority
    * - false if not
    */
   return OsSSL::peerIdentity( mSSL, altNames, commonName );
}

/* //////////////////////////// PROTECTED ///////////////////////////////// */

/* //////////////////////////// PRIVATE /////////////////////////////////// */


void OsSSLConnectionSocket::SSLInitSocket(sipx_int_t socket, sipx_long_t timeoutInSecs)
{
    if (mIsConnected)
    {
       sipx_int_t err = -1;

       // TODO: eventually this should allow for other SSL contexts...
       mSSL = OsSharedSSL::get()->getServerConnection();

       if (mSSL && (socketDescriptor > OS_INVALID_SOCKET_DESCRIPTOR))
       {
          SSL_set_fd (mSSL, socketDescriptor);

          err = SSL_connect(mSSL);
          if (err > 0)
          {
             OsSSL::logConnectParams(FAC_KERNEL, PRI_DEBUG,
                                     "OsSSLConnectionSocket",
                                     mSSL);

             // TODO: Validate certificate here? or in callback?
          }
          else
          {
             OsSSL::logError(FAC_KERNEL, PRI_ERR,
                             "OsSSLConnectionSocket SSL_connect failed: ",
                             SSL_get_error(mSSL, err));
             mIsConnected = FALSE;
             OsConnectionSocket::close();
             socketDescriptor = OS_INVALID_SOCKET_DESCRIPTOR;          
          }
       }
       else
       {
          OsSysLog::add(FAC_KERNEL, PRI_DEBUG
                        , "OsSSLConnectionSocket bad parameters mSSL=%p,%d, closing socket..."
                        , mSSL, socketDescriptor);
          mIsConnected = FALSE;
          OsConnectionSocket::close();
          socketDescriptor = OS_INVALID_SOCKET_DESCRIPTOR;
       }
    }
    else
    {
       OsSysLog::add(FAC_KERNEL, PRI_ERR,
                     "OsSSLConnectionSocket::SSLInitSocket called on unconnected socket"
                     );
    }
}
/* ============================ FUNCTIONS ================================= */
