//
// Copyright (C) 2005-2006 SIPez LLC.
// Licensed to SIPfoundry under a Contributor Agreement.
//
// Copyright (C) 2004-2006 SIPfoundry Inc.
// Licensed by SIPfoundry under the LGPL license.
//
// Copyright (C) 2004, 2005 Pingtel Corp.
// Licensed to SIPfoundry under a Contributor Agreement.
//
// $$
////////////////////////////////////////////////////////////////////////
//////

// SYSTEM INCLUDES
#include <arpa/inet.h>
#include <errno.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <net/if.h>

// APPLICATION INCLUDES
#include <os/HostAdapterAddress.h>
#include <os/OsSysLog.h>

/* Get the addresses associated with all of the IP interfaces.
 * The core work is done by the SIOCGIFCONF ioctl, documented in the
 * netdevice(7) manual page, with additional hsipx_int_ts in ioctl(2), ip(7) and inet(3).
 */
bool getAllLocalHostIps(const HostAdapterAddress* localHostAddresses[],
                        sipx_int_t &numAddresses)
{
   numAddresses = 0;
   UtlBoolean rc;

   // Allocate array of struct ifreq's.
   struct ifreq ifreq_array[MAX_IP_ADDRESSES];
   // Allocate struct ifconf.
   struct ifconf ifconf_structure;
   // Posipx_int_t ifconf to ifreq's.
   ifconf_structure.ifc_len = sizeof (ifreq_array);
   ifconf_structure.ifc_req = ifreq_array;

   // Open an arbitrary network socket on which to perform the ioctl.
   sipx_int_t sock = socket(PF_INET, SOCK_DGRAM, 0);
   if (sock < 0)
   {
      OsSysLog::add(FAC_KERNEL, PRI_ERR,
                    "getAllLocalHostIps unable to open socket, errno = %d",
                    errno);
      rc = FALSE;
   }
   else
   {
      // Perform the SIOCGIFCONF ioctl to get the interface addresses.
      sipx_int_t ret = ioctl(sock, SIOCGIFCONF, (void*) &ifconf_structure);

      if (ret < 0)
      {
         OsSysLog::add(FAC_KERNEL, PRI_ERR,
                       "getAllLocalHostIps error performing SIOCGIFCONF, errno = %d",
                       errno);
         rc = FALSE;
      }
      else
      {
         rc = TRUE;
#ifndef __MACH__
         // Get the number of returned addresses from ifc_len.
         numAddresses = ifconf_structure.ifc_len / sizeof (struct ifreq);
         sipx_int_t j = 0;
         // Iterate through the returned addresses.
         for (sipx_int_t i = 0; i < numAddresses; i++)
         {
            // Get transient pointer to address in text format.
            char* s = inet_ntoa(((struct sockaddr_in&) (ifreq_array[i].ifr_addr)).sin_addr);

            // Ignore the loopback address, because opening ports on the
            // loopback interface sipx_int_terferes with STUN operation.
            UtlString address(s);
            if (address.compareTo("127.0.0.1") != 0 && address.compareTo("0.0.0.0") != 0)
            {
               // Put the interface name and address sipx_int_to a HostAdapterAddress.
               localHostAddresses[j] = new HostAdapterAddress(ifreq_array[i].ifr_name, s);
               OsSysLog::add(FAC_KERNEL, PRI_DEBUG,
                             "getAllLocalHostIps entry %d, interface '%s', address '%s'",
                             j, ifreq_array[i].ifr_name, s);
               j++;
            }
         }
         numAddresses = j;
#else
         void* ptr;
         for (ptr = ifreq_array; (int) ptr < ((int) ifreq_array) + ifconf_structure.ifc_len; )
         {
            struct ifreq* ifr = (struct ifreq*) ptr;
            
            int len = sizeof(struct sockaddr);
            if (ifr->ifr_addr.sa_len > sizeof(struct sockaddr))
               len = ifr->ifr_addr.sa_len;

            ptr = (void *) ((int) ptr + sizeof(ifr->ifr_name) + len);

            // The body of this if statement should mirror the for loop above...
            if (ifr->ifr_addr.sa_family == AF_INET)
            {
               char* s = inet_ntoa(((struct sockaddr_in*) &ifr->ifr_addr)->sin_addr);
               UtlString address(s);
               if (address.compareTo("127.0.0.1") != 0 && address.compareTo("0.0.0.0") != 0)
               {
                  localHostAddresses[numAddresses] = new HostAdapterAddress(ifr->ifr_name, s);
                  OsSysLog::add(FAC_KERNEL, PRI_DEBUG,
                                "getAllLocalHostIps entry %d, interface '%s', address '%s'",
                                numAddresses, ifr->ifr_name, s);
                  numAddresses++;
               }
            }
         }
#endif
      }
      close(sock);
   }
   return rc;
}

bool getContactAdapterName(char* szAdapter, const char* szIp)
{
   bool found = false;
   
   UtlString ipAddress(szIp);
   
   sipx_int_t numAddresses = 0;
   const HostAdapterAddress* adapterAddresses[MAX_IP_ADDRESSES];
   getAllLocalHostIps(adapterAddresses, numAddresses);

   for (sipx_int_t i = 0; i < numAddresses; i++)
   {
      if (ipAddress.compareTo(adapterAddresses[i]->mAddress.data()) == 0)
      {
         strcpy(szAdapter, adapterAddresses[i]->mAdapter.data());
         OsSysLog::add(FAC_KERNEL, PRI_DEBUG,
                       "getContactAdapterName found name %s for ip %s",
                       szAdapter, szIp);
         found = true;
      }
      delete adapterAddresses[i];
   }
   
   if (found)
   {
      return true;
   }
   else
   {
      return false;
   }
}

