//
// Copyright (C) 2004, 2005 Pingtel Corp.
// 
//
// $$
////////////////////////////////////////////////////////////////////////

// SYSTEM INCLUDES
#include <limits.h>

// APPLICATION INCLUDES
#include "utl/UtlHistogram.h"
#include "os/OsTime.h"
#include "os/OsDateTime.h"
#include "os/OsSysLog.h"

// EXTERNAL FUNCTIONS
// EXTERNAL VARIABLES
// CONSTANTS
// STATIC VARIABLE INITIALIZATIONS

/* //////////////////////////// PUBLIC //////////////////////////////////// */

/* ============================ CREATORS ================================== */

// Constructor
UtlHistogram::UtlHistogram(sipx_uint_t bins, sipx_int_t base, sipx_uint_t size,
                           const char* outputFormat,
                           sipx_uint_t outputWidth)
{
   if (bins == 0)
   {
      OsSysLog::add(FAC_KERNEL, PRI_CRIT,
                    "UtlHistogram::UtlHistogram bins must be at least 1");
      bins = 1;
   }
   mNoOfBins = bins;

   mBase = base;

   if (size == 0)
   {
      OsSysLog::add(FAC_KERNEL, PRI_CRIT,
                    "UtlHistogram::UtlHistogram size must be at least 1");
      size = 1;
   }
   mBinSize = size;

   mpBins = new sipx_uint_t[mNoOfBins + 2];
   mCount = 0;

   mOutputFormat = outputFormat;
   mOutputWidth = outputWidth;
}

// Destructor
UtlHistogram::~UtlHistogram()
{
   delete[] mpBins;
}


/* ============================ MANIPULATORS ============================== */

sipx_uint_t UtlHistogram::tally(sipx_int_t value)
{
   // Calculate the bin number.
   sipx_uint_t bin = (value - mBase) / mBinSize;
   // Increment the appropriate bin.
   ++mpBins[bin < 0 ? 0 :
            bin > mNoOfBins ? mNoOfBins + 1 :
            bin + 1];
   // Increment the total count.
   mCount++;

   return mCount;
}

void UtlHistogram::clear()
{
   // Clear the bins.
   // Remember that there are mNoOfBins + 2 bins.
   for (sipx_uint_t i = 0; i < mNoOfBins + 2; i++)
   {
      mpBins[i] = 0;
   }
   // Clear the total count.
   mCount = 0;
}

/* ============================ ACCESSORS ================================= */


/**
 * Get the number of bins.
 */
sipx_uint_t UtlHistogram::getNoOfBins()
{
   return mNoOfBins;
}

/**
 * Get the lowest value for bin 0.
 */
sipx_uint_t UtlHistogram::getBase()
{
   return mBase;
}

/**
 * Get the size of each bin.
 */
sipx_uint_t UtlHistogram::getBinSize()
{
   return mBinSize;
}

/**
 * Get the total count.
 */
sipx_uint_t UtlHistogram::getCount()
{
   return mCount;
}

/**
 * Get the count in bin i.
 */
sipx_uint_t UtlHistogram::operator[](sipx_uint_t i)
{
   // Normalize i to index mpBins[].
   i++;
   return
      i >= 0 && i < mNoOfBins+2 ?
      mpBins[i] :
      0;
}

/* ============================ INQUIRY =================================== */

/**
 * Get a string containing the formatted values from the bins.
 */
UtlString* UtlHistogram::show()
{
   sipx_uint_t width;

   // Buffer sipx_int_to which to sprintf the values.
   char* buffer = new char[(mNoOfBins+2) * mOutputWidth + 1];

   // Translate the values.
   for (sipx_uint_t i = 0, j = 0; i < mNoOfBins + 2; i++, j += width)
   {
      width = sprintf(&buffer[j], mOutputFormat, mpBins[i]);
      if (width > mOutputWidth)
      {
         OsSysLog::add(FAC_KERNEL, PRI_CRIT,
                       "UtlHistogram::show output from format '%s' "
                       "had width %d != declared width %d",
                       mOutputFormat, width, mOutputWidth);
      }
   }
   // Since there is always at least one bin, sprintf was called at least
   // once and it ended the string with a NUL.

   // Construct the UtlString.
   UtlString* value = new UtlString(buffer);
   delete[] buffer;

   return value;
}

/* //////////////////////////// PROTECTED ///////////////////////////////// */

/* //////////////////////////// PRIVATE /////////////////////////////////// */

/* ============================ FUNCTIONS ================================= */


